from os.path import realpath

DEBUG = True
TEMPLATE_DEBUG = True

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.sessions',
    'django.contrib.sites',
    'configs',
)
SITE_ID=1
ROOT_URLCONF = 'urls'

DATABASE_NAME = realpath('cnmdb')
DATABASE_ENGINE = 'sqlite3'
TEMPLATE_DIRS = 'templates'
TEST_DATABASE_NAME = realpath('testdb')
DATABASE_SUPPORTS_TRANSACTIONS = False
DATABASE_OPTIONS = {'timeout': 30}

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
#    'django.middleware.common.CommonMiddleware',
#    'django.middleware.doc.XViewMiddleware',
)



class ExceptionBase(Exception):
    def __init__(self):
        Exception.__init__(self)
    def __str__(self):
        return self.__repr__()

class CnmServerException(ExceptionBase):
    def __init__(self, msg):
        ExceptionBase.__init__(self)
        self.msg = msg
    def __repr__(self):
        return "Error from CNM server: %s" % self.msg

class MachineStatusException(ExceptionBase):
    def __init__(self, msg):
        ExceptionBase.__init__(self)
        self.msg = msg
    def __repr__(self):
        return "Error in status data: %s" % self.msg

class InvalidAction(Exception):
    def __init__(self, package_name, action_name):
        Exception.__init__(self)
        self.package_name = package_name
        self.action_name = action_name
    def __repr__(self):
        msg = "Cannot run action %s on package %s"
        return msg % (self.action, self.package_name)
    def __str__(self):
        return self.__repr__()

class QueuedJob(ExceptionBase):
    def __init__(self, job_name):
        ExceptionBase.__init__(self)
        self.job_name = job_name
    def __repr__(self):
        return self.job_name

class CannotQueueJob(ExceptionBase):
    def __init__(self, reason):
        ExceptionBase.__init__(self)
        self.reason = reason
    def __repr__(self):
        return self.reason

class DispatcherAlreadyStarted(ExceptionBase):
    def __repr__(self):
        return ""

class DispatcherError(ExceptionBase):
    def __init__(self, explanation):
        ExceptionBase.__init__(self)
        self.explanation = explanation
    def __repr__(self):
        return "Dispatcher error: %s" % self.explanation

class InvalidDispatcherAction(ExceptionBase):
    def __init__(self, action_name):
        ExceptionBase.__init__(self)
        self.action_name = action_name
    def __repr__(self):
        return "Invalid action: %s" % self.action
    def __str__(self):
        return self.__repr__()

class JobAlreadySet(Exception):
    def __init__(self, job_name):
        Exception.__init__(self)
        self.job_name = job_name
    def __repr__(self):
        return "Another job is already running: %s" % self.job_name
    def __str__(self):
        return self.__repr__()

class EnableRequiredException(Exception):
    def __init__(self):
        Exception.__init__(self)
    def __repr__(self):
        return "Must be in enable mode to connect to this server"
    def __str__(self):
        return self.__repr__()

class IncompleteConfigurationException(Exception):
    def __init__(self, server, errmsg):
        Exception.__init__(self)
        self.server = server
        self.errmsg = errmsg
    def __repr__(self):
        msg = "Server configuration for %s is incomplete (%s)"
        return msg % (self.server, self.errmsg)
    def __str__(self):
        return self.__repr__()

class BombardierMachineException(Exception):
    def __init__(self):
        Exception.__init__(self)
    def __repr__(self):
        return "Error running bc.py"
    def __str__(self):
        return self.__repr__()

class MachineUnavailableException(Exception):
    def __init__(self, server, errmsg):
        Exception.__init__(self)
        self.server = server
        self.errmsg = errmsg
    def __repr__(self):
        return "Unable to connect to %s (%s)" % (self.server, self.errmsg)
    def __str__(self):
        return self.__repr__()

class SecureCopyException(Exception):
    def __init__(self, source, dest, errmsg):
        Exception.__init__(self)
        self.source = source
        self.dest = dest
        self.errmsg = errmsg
    def __repr__(self):
        msg = "Unable to copy (%s) to (%s): %s"
        msg = msg % (self.source, self.dest, self.errmsg)
        return msg
    def __str__(self):
        return self.__repr__()

class InvalidInput(Exception):
    def __init__(self, bad_characters):
        Exception.__init__(self)
        self.bad_characters = bad_characters
    def __repr__(self):
        return "Invalid input: %s" % (self.bad_characters)
    def __str__(self):
        return self.__repr__()

class JoinTimeout(Exception):
    def __init__(self, job_name, timeout):
        Exception.__init__(self)
        self.job_name = job_name
        self.timeout = timeout
    def __repr__(self):
        return "Timed out joining a job %s (%s)" % (self.job_name, self.timeout)
    def __str__(self):
        return self.__repr__()

class InvalidJobName(Exception):
    def __init__(self, job_name="NO_JOB_NAME"):
        Exception.__init__(self)
        self.job_name = job_name
    def __repr__(self):
        return "Requesting a job that doesn't exist: %s" % self.job_name
    def __str__(self):
        return self.__repr__()

class InvalidServerHome(Exception):
    def __init__(self, server_home):
        Exception.__init__(self)
        self.server_home = server_home
    def __repr__(self):
        return "Server home value is not valid: %s" % self.server_home
    def __str__(self):
        return self.__repr__()

class MachineConfigurationException(Exception):
    def __init__(self, server, message=''):
        self.server = server
        self.message = message
    def __repr__(self):
        msg = "Could not find valid configuration data for %s (%s)" 
        return msg % (self.server, self.message)
    def __str__(self):
        return self.__repr__()

class PackageNotFound(Exception):
    def __init__(self, package_name, path=''):
        Exception.__init__(self)
        self.package_name = package_name
        self.path = path
    def __repr__(self):
        return "Package not found: %s (%s)" % (self.package_name, self.path)
    def __str__(self):
        return self.__repr__()

class DispatcherOffline(Exception):
    def __init__(self):
        Exception.__init__(self)
    def __repr__(self):
        return "Dispatcher offline"
    def __str__(self):
        return self.__repr__()



"""This is a job that runs underneath the dispatcher. Typically is going to
perform some kind of action on a remote machine. Could also be used to
run a build on the CNM server itself"""

import ServerLogMixin
from bombardier_core.static_data import OK, FAIL
from threading import Thread
import time
import StringIO, traceback

class Job(Thread, ServerLogMixin.ServerLogMixin):
    """Job class: Runs remote commands or copies files to a remote machine.
     Watches status of job"""
    def __init__(self, username, machine_name, job_id):
        ServerLogMixin.ServerLogMixin.__init__(self)
        self.name = "%s@%s-%d" % (username, machine_name, job_id)
        self.username = username
        self.machine_name = machine_name
        self.copy_dict = {}
        self.commands = []
        self.require_status = True
        self.machine_interface = None
        self.start_time = None
        self.end_time = None
        self.command_status = OK
        self.command_output = []
        self.output_pointer = 0
        self.predecessors = []
        self.final_logs = []
        Thread.__init__(self)
        self.complete_log = ''
        self.status = OK
        self.kill_switch = False
        self.acknowledged = False
        self.description = None
        self.require_comment = False

    def info(self):
        "Tell me something about all your commands, job."
        if self.description:
            return self.description
        output = []
        for command in self.commands:
            output.append(command.info())
        return ": ".join(output)

    def record_exception(self):
        "Called when an exception is trapped on a job"
        self.command_status = FAIL
        self.command_output.append( self.dump_exception(self.username) )

    def setup(self, machine_interface, commands, copy_dict,
              require_status, predecessors):
        "Set the job up to be used"
        self.copy_dict = copy_dict
        self.commands = commands
        self.require_status = require_status
        self.machine_interface = machine_interface
        self.predecessors = predecessors

    def get_all_predecessors(self):
        "find out all the possible jobs I'm dependent on"
        all_predecessors = list(self.predecessors)
        for job in self.predecessors:
            all_predecessors += job.get_all_predecessors()
        return all_predecessors

    def __cmp__(self, other):
        """Compare myself to another job to see if I should run first
        or if the other job should run first"""
        if other in self.get_all_predecessors():
            return 1
        elif self in other.get_all_predecessors():
            return -1
        else:
            return 0

    def __repr__(self):
        "Representation of this class"
        return self.name

    def _get_elapsed_time(self):
        "Figure out how long this job took to run"
        if self.isAlive:
            if self.start_time != None:
                return time.time() - self.start_time 
            else:
                return 0
        else:
            return self.end_time - self.start_time

    def get_status_dict(self):
        "Returns the job's current status data"
        output = {}
        output["alive"] = self.isAlive()
        output["command_status"] = self.command_status
        output["job_name"] = self.name
        output["elapsed_time"] = self._get_elapsed_time()
        output["complete_log"] =  self.complete_log
        output["command_output"] = self.command_output
        if self.isAlive():
            output["new_output"] = self.machine_interface.get_new_logs()
        else:
            output["new_output"] = self.final_logs
        return output

    def terminate(self):
        "Be crusty about killing a job"
        if self.isAlive():
            self.machine_interface.terminate()
        self.end_time = time.time()

    def kill(self):
        "Nicely end a job that is currently running"
        self.machine_interface.control_c()
        self.end_time = time.time()
        self.status = FAIL

    def run(self):
        "Runs commands in command list and watches status"
        self.start_time = time.time()
        try:
            self.machine_interface.freshen(self.name, self.require_status)
            self.machine_interface.scp_dict(self.copy_dict)
            self.server_log.info("Starting...", self.name)

            for command in self.commands:
                status = self.run_command(command)
                if status != OK:
                    self.status = FAIL
                    break
            self.server_log.info("Finishing", self.name)
            self.final_logs = self.machine_interface.polling_log.get_final_logs()
            polling_log = self.machine_interface.polling_log 
            self.complete_log = polling_log.get_complete_log()
        except Exception:
            exc = StringIO.StringIO()
            traceback.print_exc(file=exc)
            exc.seek(0)
            data = exc.read()
            ermsg = ''
            for line in data.split('\n'):
                ermsg = "%% %s" % line
                self.server_log.error(ermsg, self.name)
            self.command_status = FAIL
            msg = "Exception in server. Check log for details"
            self.command_output.append(msg)
            self.status = FAIL
        self.machine_interface.unset_job()
        self.end_time = time.time()

    def run_command(self, command):
        "Run a command and get its status"
        try:
            msg = "Processing command: %s" % command.name, self.name
            self.server_log.info(msg)
            status, output = command.execute(self.machine_interface)
            self.command_status = status
            self.command_output = output
            return status
        except Exception:
            msg = "Command %s: Failed to run %s"
            msg = msg % (command.name, command.info())
            self.server_log.error( msg, self.name)
            self.command_status = FAIL
            return FAIL




"Dispatcher module"
import Pyro.core
import time, os
import re
import glob
import yaml
import Pyro.naming

from bombardier_core.static_data import OK, FAIL, WAIT
from bombardier_core.Cipher import Cipher, DecryptionException
from bombardier_core.static_data import ACTION_LOOKUP

from Exceptions import InvalidJobName, JoinTimeout
from Exceptions import QueuedJob

from MachineConfig import MachineConfig
from BombardierMachineInterface import BombardierMachineInterface
from LocalMachineInterface import LocalMachineInterface 
import Job
import ServerLogMixin
from Commands import BuildCommand, ShellCommand, BombardierCommand
import DispatchMonitor

LOCAL_MACHINE_NAME = "CNM_Server"

class Dispatcher(Pyro.core.ObjBase, ServerLogMixin.ServerLogMixin):
    "Dispatcher class, manages jobs on remote machines"
    def __init__(self):
        Pyro.core.ObjBase.__init__(self)
        ServerLogMixin.ServerLogMixin.__init__(self)
        self.start_time = time.time()
        self.password = None
        self.server_home = None
        self.next_job = 1
        self.machine_interface_pool = {}
        self.monitor = DispatchMonitor.DispatchMonitor()
        self.new_jobs = {}
        self.monitor.start()
        self.jobs_to_comment = {}

    def set_server_home(self, username, server_home):
        "Set server_home"
        self.server_log.info("Setting server home: %s" % server_home, username)
        self.server_home = server_home

    def get_machine_interface(self, username, machine_name):
        "Returns an interface to a machine, creating a new one if needed"

        if machine_name == LOCAL_MACHINE_NAME:
            machine_config = MachineConfig(LOCAL_MACHINE_NAME, self.password,
                                           self.server_home)
            machine_interface = LocalMachineInterface(machine_config,
                                                      self.server_log)
            return machine_interface 

        machine_config = MachineConfig(machine_name, self.password,
                                       self.server_home)
        machine_config.merge()
        if self.password:
            machine_config.decrypt_config()
        machine_interface = None
        if machine_name in self.machine_interface_pool:
            msg = "Reusing existing connection to %s" % machine_name
            self.server_log.info(msg, username)
            machine_interface = self.machine_interface_pool[machine_name]
            machine_interface.set_data(machine_config)
        else:
            msg = "Instantiating a MachineInterface for %s" % machine_name
            self.server_log.info(msg, username)
            machine_interface = BombardierMachineInterface(machine_config,
                                                           self.server_log)
            self.machine_interface_pool[machine_name] = machine_interface
        return machine_interface 

    def _create_next_job(self, username, machine_name):
        "Creates a job, keeping track of the master job number"
        job = Job.Job(username, machine_name, self.next_job)
        self.new_jobs[job.name] = job
        self.next_job += 1
        return job

    def _setup_job(self, job, commands, copy_dict, require_status,
                  job_predecessors = []):
        "When you want a job to run after another job is finished"
        machine_interface = self.get_machine_interface(job.username,
                                                       job.machine_name)
        job.setup(machine_interface, commands, copy_dict, require_status,
                 job_predecessors)

    def queue_job(self, job_name): 
        "Take a job and get it ready to be run"
        self.server_log.info("Getting: %s (%s)" % (job_name, type(job_name)))
        job = self.new_jobs[job_name]
        self.server_log.info("Here is the job: %s" % job)
        for predecessor_job in job.predecessors:
            if predecessor_job.name in self.monitor.broken_jobs:
                reason = "Predecessor job %s is broken."
                reason = reason % predecessor_job.name
                predecessor_job.command_status = FAIL
                predecessor_job.command_output.append(reason)
                continue
            if not predecessor_job in self.monitor.job_queue:
                if not predecessor_job.name in self.monitor.active_jobs:
                    if not predecessor_job.name in self.monitor.finished_jobs:
                        predecessor_job.command_status = FAIL
                        reason = "Cannot find predecessor job %s"
                        reason = reason % predecessor_job.name
                        predecessor_job.command_output.append(reason)
        if job.command_status == OK:
            self.monitor.job_queue.append(job)
        else:
            self.monitor.broken_jobs[job_name] = job
            self.server_log.info("BROKEN %s" % job_name)

    def get_job_status(self, job_name):
        "Find out how the job is doing"
        job = self.new_jobs[job_name]
        return job.get_status_dict()
    
    def unpush_job(self, username, machine_name):
        "remove cleartext configuration data to the remote machine for testing"
        commands = []
        job = self._create_next_job(username, machine_name)
        job.description = "Un-push"
        try:
            machine_interface = self.get_machine_interface(username,
                                                           machine_name)
            config_path = os.path.join(machine_interface.spkg_dir,
                                       machine_interface.machine_name)
            clear = ShellCommand("Removing cleartext configuration file", 
                                 "rm -f config.yml", config_path)
            commands = [clear]
            self._setup_job(job, commands, {}, False)
        except Exception:
            job.record_exception()
        return job.name

    def push_job(self, username, machine_name):
        "push cleartext configuration data to the remote machine for testing"
        job = self._create_next_job(username, machine_name)
        job.description = "Push"
        try:
            machine_interface = self.get_machine_interface(username,
                                                           machine_name)
            spkg_dir = machine_interface.spkg_dir
            src = os.path.join("machine", "%s.yml" % machine_name)
            dst = os.path.join(spkg_dir, machine_interface.machine_name,
                               "config.yml")
            copy_dict = {"config": [ src, dst ]}
            self._setup_job(job, [], copy_dict, False)
        except Exception:
            job.record_exception()
        msg = "This is the job we're making: %s (%s)"
        msg = msg % (job.name, len(job.name))
        self.server_log.info(msg)
        return job.name

    def init_job(self, username, machine_name):
        "Initialize a bombardier client install"
        job = self._create_next_job(username, machine_name)
        job.description = "Initialize"
        try:
            machine_interface = self.get_machine_interface(username,
                                                           machine_name)
            spkg_dir = machine_interface.spkg_dir
            commands = []
            if machine_interface.platform == "win32":
                rtool = "regtool"
                bom_hkey = "/HKEY_LOCAL_MACHINE/SOFTWARE/Bombardier"
                cmd = "export SPKG_DOS_DIR=$(cygpath -w %s);" % spkg_dir
                cmd += '%s add -v "%s";' % ( rtool, bom_hkey )
                cmd += '%s set -v "%s/InstallPath" $SPKG_DOS_DIR'
                cmd = cmd % ( rtool, bom_hkey )
            else:
                cmd = 'echo spkg_path: %s > /etc/bombardier.yml' % spkg_dir
            set_spkg_config = ShellCommand("Setting spkg path", cmd, '.')
            bombardier_init = BombardierCommand("init", None, None)

            commands = [set_spkg_config, bombardier_init]
            job.description = "INIT"
            job.require_comment = True
            self._setup_job(job, commands, {}, False)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def bom_job(self, username, machine_name, action_string, status_required):
        "Run a bom level job on a machine"
        job = self._create_next_job(username, machine_name)
        try:
            if action_string == 'reconcile':
                job.require_comment = True
            bombardier_recon = BombardierCommand(action_string, None, None)
            commands = [bombardier_recon]
            self._setup_job(job, commands, {}, status_required)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def check_status_job(self, username, machine_name):
        "Check a machine's install status against its bom"
        return self.bom_job(username, machine_name, "check_status",
                                              status_required=False)

    def reconcile_job(self, username, machine_name):
        "Reconciles a machine to its bom"
        return self.bom_job(username, machine_name, "reconcile",
                                            status_required=True)

    def package_build_job(self, username, package_name, svn_user,
                          svn_password, debug, prepare):
        '''
        Looks at a package .yml file and revises data regarding it
        username -- the name of the user issuing the command
        package_name -- the name of the package to deal with
        svn_user -- svn credentials
        svn_password -- svn credentials
        '''
        job = self._create_next_job(username, LOCAL_MACHINE_NAME)
        try:
            build_cmd = BuildCommand(package_name, svn_user, svn_password,
                                     debug, prepare)
            commands = [build_cmd]
            self._setup_job(job, commands, {}, True)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def package_action_job(self, username, package_name, action_string, 
                           machine_name):
        '''
        Runs a bc command for a certain machine and package
        username -- the name of the user issuing the command
        package_name -- the name of the package to deal with
        action_string -- one of the following: uninstall, configure, install,
                         verify, reconcile, check_status, execute, fix, purge,
                         dry_run, or init
        machine_name -- name of the machine to run the job on
        '''
        job = self._create_next_job(username, machine_name)
        job.require_comment = True
        script_name = ''
        if action_string not in ACTION_LOOKUP:
            script_name = action_string
            action_string = "execute"
        if package_name:
            if action_string == "check_status" or action_string == "reconcile":
                script_name = action_string
                action_string = "execute"
        try:
            bom_cmd = BombardierCommand(action_string, package_name, 
                                        script_name)
            commands = [bom_cmd]

            self._setup_job(job, commands, {}, True)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def dist_job(self, username, machine_name, dist_name):
        "Job that unpacks and installs a distutils package"
        job = self._create_next_job(username, machine_name)
        job.description = "DIST: %s" % dist_name
        job.require_comment = True
        try:
            unpack = ShellCommand("Unpacking %s on the client" % dist_name, 
                             "tar -xzf %s.tar.gz" % dist_name, "~")
            install = ShellCommand("Installing client libraries...",
                              "python setup.py install",
                              "~/%s" % dist_name)
            commands = [unpack, install]
            src_file = dist_name+".tar.gz"
            copy_dict = {"dist": [os.path.join("dist", src_file), '.']}
            self._setup_job(job, commands, copy_dict, False)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def enable_job(self, username, machine_name, password):
        "Set up ssh shared keys with a remote machine"
        job = self._create_next_job(username, machine_name)
        job.description = "Enable"
        try:
            machine_interface = self.get_machine_interface(username,
                                                           machine_name)
            machine_interface.ssh_pass = password
            public_key = "id_dsa.pub"
            copy_dict = { "admin" : [os.path.join("admin", public_key), '.'] }
            ssh_keys_dir = "~/.ssh"
            ssh_keys_dir_cmd = "[ -e %s ] || mkdir %s && chmod 700 %s" \
                             % (ssh_keys_dir, ssh_keys_dir, ssh_keys_dir)
            make_ssh_dir = ShellCommand("Creating .ssh dir if necessary",
                                        ssh_keys_dir_cmd, "~")

            cat_cmd = "cat ~/%s >> authorized_keys2" % public_key
            cat_key = ShellCommand("Adding ssh key to %s" % machine_name,
                                   cat_cmd, ssh_keys_dir)

            commands = [make_ssh_dir, cat_key]
            self._setup_job(job, commands, copy_dict, False)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def disable_job(self, username, machine_name):
        "Remove shared ssh key from a remote machine"
        job = self._create_next_job(username, machine_name)
        job.description = "Disable"
        try:
            public_key = "id_dsa.pub"
            auth_keys = ".ssh/authorized_keys2"
            tmp_auth = "tmp_auth_keys2"
            copy_dict = { "admin" : [os.path.join("admin", public_key), '.'] }

            filter_cmd = """grep -v "$( awk '{print $3}' < %s""" % public_key
            filter_cmd += """)" < %s > %s""" % ( auth_keys, tmp_auth )
            filter_sc = ShellCommand("Filtering ssh key from %s" % machine_name,
                                      filter_cmd, '~')
            overwrite_cmd = "cat %s > %s && rm %s"
            overwrite_cmd = overwrite_cmd  % ( tmp_auth, auth_keys, tmp_auth )
            overwrite_sc = ShellCommand("Overwriting authorized_keys",
                                     overwrite_cmd, '~')
            
            commands = [filter_sc, overwrite_sc]
            self._setup_job(job, commands, copy_dict, False)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def test_job(self, username, machine_name):
        "Simple test job"
        job = self._create_next_job(username, machine_name)
        job.description = "Self-test"
        try:
            cmd = 'for i in 1 2 3 4; do sleep 1; echo "Testing $i/4"; done'
            commands = [ShellCommand("self_test", cmd, '.')]
            self._setup_job(job, commands, {}, False)
        except Exception:
            job.command_output = self.dump_exception(username)
            job.command_status = FAIL
        return job.name

    def terminate(self, username):
        "Close connections for machine interfaces in interface pool"
        output = {"command_status": OK}
        try:
            self.monitor.kill_switch = True
            for machine_name in self.machine_interface_pool:
                machine_interface = self.machine_interface_pool[machine_name]
                msg = "Requested connection termination to %s" % machine_name
                self.server_log.warning(msg, username)
                status = machine_interface.terminate()
                output[machine_name] = status
            self.machine_interface_pool = {}
            self.monitor.join()
        except Exception:
            output.update(self.dump_exception(username))
        return output

    def get_jobs_pending_comments(self, username):
        "find all the jobs this user needs to comment on"
        job_info = self.monitor.get_uncommented_job_info(username)
        output = {"job_info": job_info}
        return output

    def get_machine_name(self, comment_job_name):
        "get the name of a machine, based on job name"
        return self.monitor.get_machine_name(comment_job_name)

    def note_comment(self, job_name):
        "record that the user is making a comment on a job"
        self.monitor.note_comment(job_name)

    def job_join(self, username, job_name, timeout):
        "Wait for a job to finish"
        require_comment = False
        try:
            self.server_log.info("Attempting to join %s" % job_name)
            job = self.get_job(job_name)
            if job:
                self.server_log.info("found the job")
            else:
                self.server_log.info("can't find the job")
            start_time = time.time()
            end_time = start_time + timeout
            if not job:
                raise InvalidJobName(job_name)
            while job_name in self.monitor.active_jobs:
                self.server_log.info("waiting for %s to time out" % job_name)
                if time.time() > end_time:
                    raise JoinTimeout(job_name, timeout)
                time.sleep(1)
            job.acknowledged = True
            if job.require_comment:
                require_comment = True
            self.server_log.info("finishing join %s" % job_name)
        except QueuedJob:
            output = {"command_status": WAIT,
                     }
            return output
        except Exception:
            self.server_log.info("exception in join %s" % job_name)
            output = {"command_status": FAIL,
                      "command_output": self.dump_exception(username),
                     }
            return output
        output = job.get_status_dict()
        output["jobs_requiring_comment"] = self.monitor.get_uncommented_job_info(username)
        output["children"] = self.monitor.get_job_children(job)
        return output

    def get_job(self, job_name):
        "Get a job from the appropriate bucket, whetever that is"
        job = self.monitor.active_jobs.get(job_name)
        if not job:
            job = self.monitor.finished_jobs.get(job_name)
            if not job:
                job = self.monitor.broken_jobs.get(job_name)
                if not job:
                    for job in self.monitor.job_queue:
                        if job_name == job.name:
                            predecessor = self.monitor.queue_reason(job)
                            raise QueuedJob(predecessor)
                    raise InvalidJobName(job_name)
        return job

    def job_poll(self, username, job_names):
        "Poll a job for status info and log data"
        job_status_dict = {}
        for job_name in job_names:
            try:
                job = self.get_job(job_name)
                job_status_dict[job_name] = job.get_status_dict()
            except QueuedJob, qj:
                job_status_dict[job_name] = {"command_status": "QUEUED",
                                             "pending_job": qj.job_name}
        return job_status_dict

    def job_kill(self, username, job_name):
        "Be mean to a job and get rid of it"
        status = FAIL
        self.server_log.info("%s killing job %s" % (username, job_name))
        try:
            status = self.monitor.kill_job(job_name)
        except Exception:
            return {"command_status": FAIL}
        return {"command_status": status}

    def clear_broken(self, username, machine_name):
        "clear out the broken dictionary for a given machine"
        jobs_cleared = self.monitor.clear_broken(machine_name)
        command_output = []
        for job_name in jobs_cleared:
            command_output.append("%s cleared" % job_name)
        output = {"command_output": command_output,
                  "command_status": OK
                 }
        return output

    def show_jobs(self, username, machine_name):
        "show detailed job information for a machine"
        job_info = self.monitor.show_job_info(machine_name)
        output = {"command_output": '',
                  "command_status": OK,
                  "job_info": job_info,
                 }
        return output

    def stop_all_jobs(self, username, machine_name):
        "stop all jobs for a given machine"
        jobs_killed, jobs_removed = self.monitor.kill_jobs(machine_name)
        command_output = []
        for job_name in jobs_killed:
            command_output.append("%s killed" % job_name)
        for job_name in jobs_removed:
            command_output.append("%s removed" % job_name)
        output = {"command_output": command_output,
                  "command_status": OK
                 }
        return output

    def check_status(self):
        "Return elapsed time"
        time_running = time.time() - self.start_time
        return { "uptime": time_running,
                 "job_queue": str(self.monitor.job_queue),
                 "active_jobs": self.monitor.active_jobs.keys(),
                 "finished_jobs": self.monitor.finished_jobs.keys(),
                 "broken_jobs": self.monitor.broken_jobs.keys(),
               }

    def set_password(self, password):
        "Set password for configuration item encryption"
        status, validate_output = self._validate_password(password)
        if status == OK:
            self.server_log.info("Setting configuration password.")
            self.password = password
            for machine_name in self.machine_interface_pool:
                machine_config = MachineConfig(machine_name, self.password,
                                               self.server_home)
                machine_config.merge()
                machine_config.decrypt_config()
                machine_interface = self.machine_interface_pool[machine_name]
                self.server_log.info("Decrypting config for %s" % machine_name)
                machine_interface.set_data(machine_config)

        output = {"command_status": status,
                  "command_output": validate_output}
        return output

    def check_password(self):
        "Returns configuration password"
        return self.password

    def _validate_password(self, password):
        "Very shallow password validation"
        lazy_dog = "the_quick_brown_fox_jumped_over_the_lazy_dog\n"
        test_decrypt_file = os.path.join(self.server_home, 'admin',
                                         'encryption_validation.yml')
        if not os.path.isfile(test_decrypt_file):
            msg = "%s doesn't exist, creating..." % test_decrypt_file
            self.server_log.warning( msg )
            cipher = Cipher(password)
            enc_lazy = cipher.encrypt_string(lazy_dog)
            enc_dict = { "enc_test" : enc_lazy }
            open( test_decrypt_file, 'w' ).write(yaml.dump( enc_dict ))
        try:
            cipher_dict = yaml.load(open(test_decrypt_file, 'r').read())
        except IOError:
            return FAIL, "Encryption not set up properly. %s not readable"
        try:
            cipher = Cipher(password)
            clear_dict = cipher.decrypt_dict(cipher_dict)
        except DecryptionException:
            return FAIL, ["Invalid configuration key."]
        if clear_dict.get("test") == lazy_dog:
            return OK, ['Configuration key set.']
        return FAIL, ["Invalid configuration key."]

    def _find_latest_file(self, matcher):
        "Looks in the dist directory for the newest version of a file"
        dist_glob_path = os.path.join(self.server_home, "dist", "*.tar.gz")
        latest_filename = ''
        max_version = 0
        for filename in glob.glob(dist_glob_path):
            matches = matcher.findall(filename)
            if matches:
                new_version = int(matches[0])
                if new_version > max_version:
                    latest_filename = filename
                    max_version = new_version
        return latest_filename

    def _dist_install_job(self, search_string, username, machine_name):
        """Find the latest version of a distribution file and create a job to
        deploy it on a machine"""
        matcher = re.compile(search_string)
        newest_dist = self._find_latest_file(matcher)
        newest_dist = newest_dist.split('.tar.gz')[0]
        newest_dist = newest_dist.split(os.path.sep)[-1]
        job_name = self.dist_job(username, machine_name, newest_dist)
        return job_name

    def setup_machine(self, username, machine_name, password):
        """Sets up a machine. This requires
           (1) enable
           (2) set up the Bombardier client/core libraries
           (3) Install PyYAML
           (4) run an 'init' job
           (5) run a 'check_status' job
        """
        job1_name = self.enable_job(username, machine_name, password)
        job1 = self.new_jobs[job1_name]

        dist_name = "bombardier_core\-1\.00\-(\d+)\.tar\.gz"
        job2_name = self._dist_install_job(dist_name, username, machine_name)
        job2 = self.new_jobs[job2_name]
        job2.predecessors = [job1]

        dist_name = "bombardier_client\-1\.00\-(\d+)\.tar\.gz"
        job3_name = self._dist_install_job(dist_name, username, machine_name)
        job3 = self.new_jobs[job3_name]
        job3.predecessors = [job2]

        dist_name = "PyYAML-3.(\d+).tar.gz"
        job4_name = self._dist_install_job(dist_name, username, machine_name)
        job4 = self.new_jobs[job4_name]
        job4.predecessors = [job3]

        job5_name = self.init_job(username, machine_name)
        job5 = self.new_jobs[job5_name]
        job5.predecessors = [job4]

        job6_name = self.package_action_job(username, '', "check_status",
                                            machine_name)
        job6 = self.new_jobs[job6_name]
        job6.predecessors = [job5]
        job_names = [job1_name, job2_name, job3_name, job4_name,
                     job5_name, job6_name]
        for job_name in job_names:
            self.queue_job(job_name)
        return job1_name

if __name__ == '__main__':
    from django.core.management import setup_environ
    import settings
    setup_environ(settings)

    Pyro.core.initServer()
    daemon = Pyro.core.Daemon()
#    ns = Pyro.naming.NameServerLocator().getNS()
#    daemon.useNameServer(ns)
#
#    #daemon=Pyro.core.Daemon()
    uri = daemon.connect(Dispatcher(), "dispatcher")
    print "The daemon runs on port:", daemon.port
    print "The object's uri is:", uri
#
    print "Started server."
    daemon.requestLoop()
#



"""
Data format classes ("responders") that can be plugged 
into model_resource.ModelResource and determine how
the objects of a ModelResource instance are rendered
(e.g. serialized to XML, rendered by templates, ...).
"""
from django.core import serializers
from django.core.handlers.wsgi import STATUS_CODE_TEXT
from django.core.paginator import QuerySetPaginator, InvalidPage
# the correct paginator for Model objects is the QuerySetPaginator,
# not the Paginator! (see Django doc)
from django.core.xheaders import populate_xheaders
from django import forms
from django.http import Http404, HttpResponse
from django.forms.util import ErrorDict
from django.shortcuts import render_to_response
from django.template import loader, RequestContext
from django.utils import simplejson
from django.utils.xmlutils import SimplerXMLGenerator
from django.views.generic.simple import direct_to_template
import os
import syck

class JsonDictResponder(object):

    def __init__(self, data):
        self.data = data

    def element(self, request):
        response = HttpResponse(mimetype = "application/json")
        simplejson.dump(self.data, response)
        return response

    def error(self, request, status_code, error_dict=None):
        raise Http404

    def list(self, request, queryset, page=None):
        raise Http404


class YamlFileResponder(object):

    def __init__(self, base_path):
        self.base_path = base_path

    def element(self, request, elem):
        full_file_path = os.path.join(self.base_path, elem+".yml")
        data = open(full_file_path).read()
        response = HttpResponse(mimetype = "application/json")
        response_dict = syck.load(data)
        simplejson.dump(response_dict, response)
        return response

    def error(self, request, status_code, error_dict=None):
        raise Http404

    def list(self, request, queryset, page=None):
        raise Http404


class SerializeResponder(object):
    """
    Class for all data formats that are possible
    with Django's serializer framework.
    """
    def __init__(self, format, mimetype=None, paginate_by=None, allow_empty=False):
        """
        format:
            may be every format that works with Django's serializer
            framework. By default: xml, python, json, (yaml).
        mimetype:
            if the default None is not changed, any HttpResponse calls 
            use settings.DEFAULT_CONTENT_TYPE and settings.DEFAULT_CHARSET
        paginate_by:
            Number of elements per page. Default: All elements.
        """
        self.format = format
        self.mimetype = mimetype
        self.paginate_by = paginate_by
        self.allow_empty = allow_empty
        self.expose_fields = []
        self.expose_all_fields = False

    def render(self, object_list):
        """
        Serializes a queryset to the format specified in
        self.format.
        """
        # Hide unexposed fields
        hidden_fields = []
        if not self.expose_all_fields:
            for obj in list(object_list):
                for field in obj._meta.fields:
                    if not field.name in self.expose_fields and field.serialize:
                        field.serialize = False
                        hidden_fields.append(field)
        response = serializers.serialize(self.format, object_list)
        # Show unexposed fields again
        for field in hidden_fields:
            field.serialize = True
        return response

    def element(self, request, elem):
        """
        Renders single model objects to HttpResponse.
        """
        return HttpResponse(self.render([elem]), self.mimetype)

    def error(self, request, status_code, error_dict=None):
        """
        Handles errors in a RESTful way.
        - appropriate status code
        - appropriate mimetype
        - human-readable error message
        """
        if not error_dict:
            error_dict = ErrorDict()
        response = HttpResponse(mimetype = self.mimetype)
        response.write('%d %s' % (status_code, STATUS_CODE_TEXT[status_code]))
        if error_dict:
            response.write('\n\nErrors:\n')
            response.write(error_dict.as_text())
        response.status_code = status_code
        return response
    
    def list(self, request, queryset, page=None):
        """
        Renders a list of model objects to HttpResponse.
        """
        if self.paginate_by:
            paginator = QuerySetPaginator(queryset, self.paginate_by)
            if not page:
                page = request.GET.get('page', 1)
            try:
                page = int(page)
                object_list = paginator.page(page).object_list
            except (InvalidPage, ValueError):
                if page == 1 and self.allow_empty:
                    object_list = []
                else:
                    return self.error(request, 404)
        else:
            object_list = list(queryset)
        return HttpResponse(self.render(object_list), self.mimetype)
    
class JSONResponder(SerializeResponder):
    """
    JSON data format class.
    """
    def __init__(self, paginate_by=None, allow_empty=False):
        SerializeResponder.__init__(self, 'json', 'application/json',
                    paginate_by=paginate_by, allow_empty=allow_empty)

    def error(self, request, status_code, error_dict=None):
        """
        Return JSON error response that includes a human readable error
        message, application-specific errors and a machine readable
        status code.
        """
        if not error_dict:
            error_dict = ErrorDict()
        response = HttpResponse(mimetype = self.mimetype)
        response.status_code = status_code
        response_dict = {
            "error-message" : '%d %s' % (status_code, STATUS_CODE_TEXT[status_code]),
            "status-code" : status_code,
            "model-errors" : error_dict.as_ul()
        }
        simplejson.dump(response_dict, response)
        return response

class XMLResponder(SerializeResponder):
    """
    XML data format class.
    """
    def __init__(self, paginate_by=None, allow_empty=False):
        SerializeResponder.__init__(self, 'xml', 'application/xml',
                    paginate_by=paginate_by, allow_empty=allow_empty)

    def error(self, request, status_code, error_dict=None):
        """
        Return XML error response that includes a human readable error
        message, application-specific errors and a machine readable
        status code.
        """
        from django.conf import settings
        if not error_dict:
            error_dict = ErrorDict()
        response = HttpResponse(mimetype = self.mimetype)
        response.status_code = status_code
        xml = SimplerXMLGenerator(response, settings.DEFAULT_CHARSET)
        xml.startDocument()
        xml.startElement("django-error", {})
        xml.addQuickElement(name="error-message", contents='%d %s' % (status_code, STATUS_CODE_TEXT[status_code]))
        xml.addQuickElement(name="status-code", contents=str(status_code))
        if error_dict:
            xml.startElement("model-errors", {})
            for (model_field, errors) in error_dict.items():
                for error in errors:
                    xml.addQuickElement(name=model_field, contents=error)
            xml.endElement("model-errors")
        xml.endElement("django-error")
        xml.endDocument()
        return response

class TemplateResponder(object):
    """
    Data format class that uses templates (similar to Django's
    generic views).
    """
    def __init__(self, template_dir, paginate_by=None, template_loader=loader,
                 extra_context=None, allow_empty=False, context_processors=None,
                 template_object_name='object', mimetype=None):
        self.template_dir = template_dir
        self.paginate_by = paginate_by
        self.template_loader = template_loader
        if not extra_context:
            extra_context = {}
        for key, value in extra_context.items():
            if callable(value):
                extra_context[key] = value()
        self.extra_context = extra_context
        self.allow_empty = allow_empty
        self.context_processors = context_processors
        self.template_object_name = template_object_name
        self.mimetype = mimetype
        self.expose_fields = None # Set by Collection.__init__
            
    def _hide_unexposed_fields(self, obj, allowed_fields):
        """
        Remove fields from a model that should not be public.
        """
        for field in obj._meta.fields:
            if not field.name in allowed_fields and \
               not field.name + '_id' in allowed_fields:
                obj.__dict__.pop(field.name)    

    def list(self, request, queryset, page=None):
        """
        Renders a list of model objects to HttpResponse.
        """
        template_name = '%s/%s_list.html' % (self.template_dir, queryset.model._meta.module_name)
        if self.paginate_by:
            paginator = QuerySetPaginator(queryset, self.paginate_by)
            if not page:
                page = request.GET.get('page', 1)
            try:
                page = int(page)
                object_list = paginator.page(page).object_list
            except (InvalidPage, ValueError):
                if page == 1 and self.allow_empty:
                    object_list = []
                else:
                    raise Http404
            current_page = paginator.page(page)
            c = RequestContext(request, {
                '%s_list' % self.template_object_name: object_list,
                'is_paginated': paginator.num_pages > 1,
                'results_per_page': self.paginate_by,
                'has_next': current_page.has_next(),
                'has_previous': current_page.has_previous(),
                'page': page,
                'next': page + 1,
                'previous': page - 1,
                'last_on_page': current_page.end_index(),
                'first_on_page': current_page.start_index(),
                'pages': paginator.num_pages,
                'hits' : paginator.count,
            }, self.context_processors)
        else:
            object_list = queryset
            c = RequestContext(request, {
                '%s_list' % self.template_object_name: object_list,
                'is_paginated': False
            }, self.context_processors)
            if not self.allow_empty and len(queryset) == 0:
                raise Http404
        # Hide unexposed fields
        for obj in object_list:
            self._hide_unexposed_fields(obj, self.expose_fields)
        c.update(self.extra_context)        
        t = self.template_loader.get_template(template_name)
        return HttpResponse(t.render(c), mimetype=self.mimetype)

    def element(self, request, elem):
        """
        Renders single model objects to HttpResponse.
        """
        template_name = '%s/%s_detail.html' % (self.template_dir, elem._meta.module_name)
        t = self.template_loader.get_template(template_name)
        c = RequestContext(request, {
            self.template_object_name : elem,
        }, self.context_processors)
        # Hide unexposed fields
        self._hide_unexposed_fields(elem, self.expose_fields)
        c.update(self.extra_context)
        response = HttpResponse(t.render(c), mimetype=self.mimetype)
        populate_xheaders(request, response, elem.__class__, getattr(elem, elem._meta.pk.name))
        return response
    
    def error(self, request, status_code, error_dict=None):
        """
        Renders error template (template name: error status code).
        """
        if not error_dict:
            error_dict = ErrorDict()
        response = direct_to_template(request, 
            template = '%s/%s.html' % (self.template_dir, str(status_code)),
            extra_context = { 'errors' : error_dict },
            mimetype = self.mimetype)
        response.status_code = status_code
        return response
    
    def create_form(self, request, queryset, form_class):
        """
        Render form for creation of new collection entry.
        """
        ResourceForm = forms.form_for_model(queryset.model, form=form_class)
        if request.POST:
            form = ResourceForm(request.POST)
        else:
            form = ResourceForm()
        template_name = '%s/%s_form.html' % (self.template_dir, queryset.model._meta.module_name)
        return render_to_response(template_name, {'form':form})

    def update_form(self, request, pk, queryset, form_class):
        """
        Render edit form for single entry.
        """
        # Remove queryset cache by cloning the queryset
        queryset = queryset._clone()
        elem = queryset.get(**{queryset.model._meta.pk.name : pk})
        ResourceForm = forms.form_for_instance(elem, form=form_class)
        if request.PUT:
            form = ResourceForm(request.PUT)
        else:
            form = ResourceForm()
        template_name = '%s/%s_form.html' % (self.template_dir, elem._meta.module_name)
        return render_to_response(template_name, 
                {'form':form, 'update':True, self.template_object_name:elem})



from django.http import HttpResponse
from django.utils.translation import ugettext as _
import md5, time, random

def djangouser_auth(username, password):
    """
    Check username and password against
    django.contrib.auth.models.User
    """
    from django.contrib.auth.models import User
    try:
        user = User.objects.get(username=username)
        if user.check_password(password):
            return True
        else:
            return False
    except User.DoesNotExist:
        return False

class NoAuthentication(object):
    """
    No authentication: Permit every request.
    """
    def is_authenticated(self, request):
        return True

    def challenge_headers(self):
        return {}

class HttpBasicAuthentication(object):
    """
    HTTP/1.0 basic authentication.
    """    
    def __init__(self, authfunc=djangouser_auth, realm=_('Restricted Access')):
        """
        authfunc:
            A user-defined function which takes a username and
            password as its first and second arguments respectively
            and returns True if the user is authenticated
        realm:
            An identifier for the authority that is requesting
            authorization
        """
        self.realm = realm
        self.authfunc = authfunc
    
    def challenge_headers(self):
        """
        Returns the http headers that ask for appropriate
        authorization.
        """
        return {'WWW-Authenticate' : 'Basic realm="%s"' % self.realm}
    
    def is_authenticated(self, request):
        """
        Checks whether a request comes from an authorized user.
        """
        if not request.META.has_key('HTTP_AUTHORIZATION'):
            return False
        (authmeth, auth) = request.META['HTTP_AUTHORIZATION'].split(' ', 1)
        if authmeth.lower() != 'basic':
            return False
        auth = auth.strip().decode('base64')
        username, password = auth.split(':', 1)
        return self.authfunc(username=username, password=password)

def digest_password(realm, username, password):
    """
    Construct the appropriate hashcode needed for HTTP digest
    """
    return md5.md5("%s:%s:%s" % (username, realm, password)).hexdigest()

class HttpDigestAuthentication(object):
    """
    HTTP/1.1 digest authentication (RFC 2617).
    Uses code from the Python Paste Project (MIT Licence).
    """    
    def __init__(self, authfunc, realm=_('Restricted Access')):
        """
        authfunc:
            A user-defined function which takes a username and
            a realm as its first and second arguments respectively
            and returns the combined md5 hash of username,
            authentication realm and password.
        realm:
            An identifier for the authority that is requesting
            authorization
        """
        self.realm = realm
        self.authfunc = authfunc
        self.nonce    = {} # prevention of replay attacks

    def get_auth_dict(self, auth_string):
        """
        Splits WWW-Authenticate and HTTP_AUTHORIZATION strings
        into a dictionaries, e.g.
        {
            nonce  : "951abe58eddbb49c1ed77a3a5fb5fc2e"',
            opaque : "34de40e4f2e4f4eda2a3952fd2abab16"',
            realm  : "realm1"',
            qop    : "auth"'
        }
        """
        amap = {}
        for itm in auth_string.split(", "):
            (k, v) = [s.strip() for s in itm.split("=", 1)]
            amap[k] = v.replace('"', '')
        return amap

    def get_auth_response(self, http_method, fullpath, username, nonce, realm, qop, cnonce, nc):
        """
        Returns the server-computed digest response key.
        
        http_method:
            The request method, e.g. GET
        username:
            The user to be authenticated
        fullpath:
            The absolute URI to be accessed by the user
        nonce:
            A server-specified data string which should be 
            uniquely generated each time a 401 response is made
        realm:
            A string to be displayed to users so they know which 
            username and password to use
        qop:
            Indicates the "quality of protection" values supported 
            by the server.  The value "auth" indicates authentication.
        cnonce:
            An opaque quoted string value provided by the client 
            and used by both client and server to avoid chosen 
            plaintext attacks, to provide mutual authentication, 
            and to provide some message integrity protection.
        nc:
            Hexadecimal request counter
        """
        ha1 = self.authfunc(realm, username)
        ha2 = md5.md5('%s:%s' % (http_method, fullpath)).hexdigest()
        if qop:
            chk = "%s:%s:%s:%s:%s:%s" % (ha1, nonce, nc, cnonce, qop, ha2)
        else:
            chk = "%s:%s:%s" % (ha1, nonce, ha2)
        computed_response = md5.md5(chk).hexdigest()
        return computed_response
    
    def challenge_headers(self, stale=''):
        """
        Returns the http headers that ask for appropriate
        authorization.
        """
        nonce  = md5.md5(
            "%s:%s" % (time.time(), random.random())).hexdigest()
        opaque = md5.md5(
            "%s:%s" % (time.time(), random.random())).hexdigest()
        self.nonce[nonce] = None
        parts = {'realm': self.realm, 'qop': 'auth',
                 'nonce': nonce, 'opaque': opaque }
        if stale:
            parts['stale'] = 'true'
        head = ", ".join(['%s="%s"' % (k, v) for (k, v) in parts.items()])
        return {'WWW-Authenticate':'Digest %s' % head}
    
    def is_authenticated(self, request):
        """
        Checks whether a request comes from an authorized user.
        """
        
        # Make sure the request is a valid HttpDigest request
        if not request.META.has_key('HTTP_AUTHORIZATION'):
            return False
        fullpath = request.META['SCRIPT_NAME'] + request.META['PATH_INFO']
        (authmeth, auth) = request.META['HTTP_AUTHORIZATION'].split(" ", 1)
        if authmeth.lower() != 'digest':
            return False
        
        # Extract auth parameters from request
        amap = self.get_auth_dict(auth)
        try:
            username = amap['username']
            authpath = amap['uri']
            nonce    = amap['nonce']
            realm    = amap['realm']
            response = amap['response']
            assert authpath.split("?", 1)[0] in fullpath
            assert realm == self.realm
            qop      = amap.get('qop', '')
            cnonce   = amap.get('cnonce', '')
            nc       = amap.get('nc', '00000000')
            if qop:
                assert 'auth' == qop
                assert nonce and nc
        except:
            return False

        # Compute response key    
        computed_response = self.get_auth_response(request.method, fullpath, username, nonce, realm, qop, cnonce, nc)
        
        # Compare server-side key with key from client
        # Prevent replay attacks
        if not computed_response or computed_response != response:
            if nonce in self.nonce:
                del self.nonce[nonce]
            return False
        pnc = self.nonce.get(nonce,'00000000')
        if nc <= pnc:
            if nonce in self.nonce:
                del self.nonce[nonce]
            return False # stale = True
        self.nonce[nonce] = nc
        return True
    



"""A job runs one or more commands"""

from bombardier_core.static_data import ACTION_LOOKUP
from bombardier_core.static_data import ACTION_REVERSE_LOOKUP
from Exceptions import InvalidAction

PENDING = 4

class AbstractCommand:
    "Abstract class for a dispatched command"
    def __init__(self, name):
        self.name = name
        self.dump_config = False
        self.status = PENDING

    def execute(self, machine_interface):
        "Virtual function for executing command"
        pass

    def info(self):
        "Default info function"
        return self.name


class BuildCommand(AbstractCommand):
    "Package build command"
    def __init__(self, package_name, svn_user, svn_password,
                 debug, prepare):
        name = "Build-%s" % (package_name)
        AbstractCommand.__init__(self, name)
        self.package_name = package_name
        self.svn_user = svn_user
        self.svn_password = svn_password
        self.debug = debug
        self.prepare = prepare

    def execute(self, machine_interface):
        "Build the package"
        return machine_interface.build_components(self.package_name,
                                                  self.svn_user,
                                                  self.svn_password,
                                                  self.debug, self.prepare)

    def info(self):
        "Return command to be run on the remote machine"
        return "Build package %s" % self.package_name

class ShellCommand(AbstractCommand):
    "Remote shell command"
    def __init__(self, name, cmd, working_dir):
        AbstractCommand.__init__(self, name)
        self.working_dir = working_dir
        self.cmd = cmd

    def execute(self, machine_interface):
        "Run the command on a remote machine"
        machine_interface.chdir(self.working_dir)
        return machine_interface.run_cmd(self.cmd)

    def info(self):
        "Return command to be run on the remote machine"
        return self.cmd

class BombardierCommand(AbstractCommand):
    """Bombardier command is a command running bc.py for 
       package or status actions"""
    def __init__(self, action_string, package_name, script_name):
        '''
        action_string -- one of the following: uninstall, configure, install,
                         verify, reconcile, check_status, execute, fix, purge,
                         dry_run, or init
        package_name -- the name of a package to act upon
        script_name -- the name of a script to act upon
        debug -- defunct, apparently
        '''
        action_const = ACTION_LOOKUP.get(action_string.lower().strip())
        if action_const == None:
            raise InvalidAction(package_name, action_string)
        name = "Bombardier-%s-%s" % (action_string, package_name)
        AbstractCommand.__init__(self, name)
        self.action = action_const
        self.package_name = package_name
        self.script_name = script_name
        self.debug = False

    def execute(self, machine_interface):
        "Run bc command"
        return machine_interface.take_action(self.action, self.package_name,
                                             self.script_name, self.debug)

    def info(self):
        "Return information about this bombardier command"
        output = ACTION_REVERSE_LOOKUP[self.action]
        if self.package_name:
            output += ": %s" % self.package_name
        if self.script_name:
            output += " (%s)" % self.script_name
        return output





class MockDict:
    """In order to instantiate a target installer package and discover what
    configuration items it might need, we construct a dictionary-like object
    that records all of the queries that is made to it"""
    def __init__(self, name):
        "Create the basic mock dictionary"
        self.name = name
        self.dicts = []

    def get(self, key):
        "Allows for treating the object like a dictionary"
        md = MockDict(key)
        self.dicts.append(md)
        return md

    def get_requests(self):
        "Divulges all of the requests that have been made to this object"
        output = {}
        for subdict in self.dicts:
            key = subdict.name
            if subdict.dicts:
                value = subdict.get_requests()
            else:
                value = ""
            output[key] = value
        return output

def parse_section(mydict, section_string, default=''):
    """This routine provides for data to be accessed in "dot-syntax" rather
    than using brackets (e.g. a dictionary such as obj = {'foo': {'bar: 1}}
    can be accessed with obj("foo.bar") rather than obj["foo"]["bar"] )"""

    sections = section_string.split('.')
    current_section = sections[0]
    if len(sections) > 1:
        if not mydict.has_key(current_section):
            mydict[current_section] = {}
        remaining_section_string = '.'.join(sections[1:])
        mydict[current_section] = parse_section(mydict[current_section],
                                                remaining_section_string,
                                                default)
    else:
        mydict[current_section] = default
    return mydict

class MockConfig:
    """This object mocks the bombardier.Config object in order to discover
    what configuration items are being requested"""

    def __init__(self):
        self.requests = {}
        self.data = MockDict("/")

    def string(self, section, default='', optional=False):
        """Obtain a string object from the configuration. Note that if the
        item we're looking for is OPTIONAL, then we don't record it."""
        if not optional:
            self.requests = parse_section(self.requests, section, default)
        return default

    def integer(self, section, default='', optional=False):
        "Obtain an integer from the configuration"
        if not optional:
            self.requests = parse_section(self.requests, section, default)
        return default

    def dictionary(self, section, default={}, optional=False):
        "Obtain a dictionary object from the configuration"
        if not optional:
            self.requests = parse_section(self.requests, section, default)
        return default

    def listobj(self, section, default=[], optional=False):
        "obtain a list object from the configuration"
        if not optional:
            self.requests = parse_section(self.requests, section, [])
        return default

    def get(self, section, option, default="", optional=False):
        "Allow this object to be treated like a dictionary"
        if not self.requests.has_key(section):
            self.requests[section] = {}
        if not optional:
            self.requests[section][option] = default
        return default

    def get_requests(self):
        """This is where the mock object tattles on what information was
        requested"""
        requests = self.requests
        data_requests = self.data.get_requests()
        for key in data_requests.keys():
            requests[key] = data_requests[key]
        return requests

    def get_instance(self):
        """Some packages want to know what instance they are running on.
        This is implemented in a trivial way in the mock system."""
        return "BUILD"





import bombardier_server.web.cnm.ServerConfig as ServerConfig
import django.utils.simplejson as json
import os


def create_server_dir(s):
    print "creating"
    os.system("mkdir -p %s" %s.server_home)
    for dir in ["include", "client", "bom"]:
        os.system("mkdir %s/%s" % ( s.server_home, dir ))

def clean_server_dir(s):
    os.system("mkdir -p %s" %s.server_home)

def setup(self):
    new_initial = [{"pk": 1, "model": "configs.include", "fields": {"name": "otherapp"}}, {"pk": 2, "model": "configs.include", "fields": {"name": "app1"}}, {"pk": 1, "model": "configs.client", "fields": {"name": "tester1"}}, {"pk": 2, "model": "configs.client", "fields": {"name": "other1"}}, {"pk": 3, "model": "configs.client", "fields": {"name": "tester2"}}, {"pk": 1, "model": "configs.bom", "fields": {"name": "foo"}}, {"pk": 2, "model": "configs.bom", "fields": {"name": "bomp"}}]
    s = ServerConfig.ServerConfig()
    create_server_dir(s)
    home = s.server_home
    for i in new_initial:
        config_file = "%s/%s/%s.yml" % ( s.server_home, i['model'].split('.')[1], i['fields']['name'] )
        print config_file
        open( config_file, 'w').write( json.dumps(i["fields"]) )

setup(1)




"Authentication urls module"
from django.conf.urls.defaults import patterns, url
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from CnmResource import CnmResource
from django.contrib.auth.views import login, logout
from django_restapi.responder import JsonDictResponder
from django_restapi.responder import JSONResponder

from bombardier_core.static_data import OK, FAIL
from django.http import Http404
import yaml

class AuthEntry(CnmResource):
    "Authentication entry"
    def read(self, request):
        "Reject read if user is not authenticated"
        if not request.user.is_authenticated():
            status = "NOT_AUTHENTICATED"
            status_code = 403
        else:
            status = "OK"
            status_code = 200
        responder = JsonDictResponder({"status": status})
        response = responder.element(request)
        response.status_code = status_code
        return response

class UserProfileEntry(CnmResource):
    "User profile entry"
    def read(self, request):
        "Return user name and superuser status as json"
        output = {"username": request.user.username,
                  "super_user": request.user.is_superuser}

        responder = JsonDictResponder(output)
        return responder.element(request)

class UsersCollection(CnmResource):
    "Users collection"
    @login_required
    def read(self, request, username):
        "Return list of possible users as json"
        users = User.objects.filter(username__startswith=username)
        responder = JSONResponder()
        responder.expose_fields = ["username"]
        return responder.list(request, users)

class UsersEntry(CnmResource):
    "Users entry"
    @login_required
    def read(self, request, username):
        "Return user info as json if user exists"
        try:
            user = User.objects.get(username__exact=username)
        except User.DoesNotExist:
            raise Http404
        output = {"username": user.username,
                  "is_superuser": user.is_superuser,
                  "is_active": user.is_active,
                  "first_name": user.first_name,
                  "last_name": user.last_name,
                  "email": user.email,
                 }
        responder = JsonDictResponder(output)
        response = responder.element(request)
        response.status_code = 200
        return response

    @login_required
    def delete(self, request, username):
        "Choose to delete a user"

        command_output = []
        user = User.objects.get(username__exact=username)
        user.delete()
        output = "%s deleted" % username
        responder = JsonDictResponder({"command_status": OK,
                                       "command_output": output})
        response = responder.element(request)
        response.status_code = 200
        return response

    @login_required
    def create(self, request, username):
        "Create or update a user"

        query_dict = {}
        command_output = []
        if "yaml" in request.POST:
            yaml_string = request.POST["yaml"]
            try:
                query_dict = yaml.load(yaml_string)
            except Exception, x:
                output = {"command_status": FAIL, "command_output": "Bad YAML"}
                responder = JsonDictResponder(output)
                return responder.element(request)
        else:
            query_dict = request.POST

        try:
            user = User.objects.get(username__exact=username)
        except User.DoesNotExist:
            user = User.objects.create_user(username,
                                            query_dict.get("email", ''))
            command_output.append("User %s created" % username)

        password = query_dict.get("password")
        if password:
            user.set_password(password)
            command_output.append("Password set for user %s" % username)

        if query_dict.has_key('admin'):
            user.is_superuser = query_dict.get('admin')
            command_output.append("User %s made an administrator" % username)
        if query_dict.has_key('active'):
            user.is_active = query_dict.get('active')
            command_output.append("User %s made active" % username)

        first_name = query_dict.get("first_name")
        if first_name:
            user.first_name = first_name
            command_output.append("First name set for user %s" % username)

        last_name = query_dict.get("last_name")
        if last_name:
            user.last_name = last_name
            command_output.append("Last name set for user %s" % username)

        email = query_dict.get("email")
        if email:
            user.email = email
            command_output.append("Email set for user %s" % username)

        user.save()
        output_string = '|'.join(command_output)
        responder = JsonDictResponder({"command_status": OK,
                                       "command_output": output_string})
        response = responder.element(request)
        response.status_code = 200
        return response

urlpatterns = patterns('',
   url(r'^json/check_authentication', AuthEntry(permitted_methods=['GET'])),
   url(r'^json/user/search/(?P<username>.*)', UsersCollection()),
   url(r'^json/user/name/(?P<username>.*)',
       UsersEntry(permitted_methods=['POST', 'GET', 'DELETE'])),
   url(r'^accounts/login/$',  login),
   url(r'^accounts/logout/$', logout),
   url(r'^accounts/profile/$', UserProfileEntry()),
)




#!/usr/bin/env python
"MachineInterface: pexpect machine interface module"

import pxssh, pexpect
import time, os, re
from bombardier_core.static_data import OK, FAIL
from Exceptions import EnableRequiredException
from Exceptions import IncompleteConfigurationException
from Exceptions import MachineUnavailableException
from Exceptions import SecureCopyException
from AbstractMachineInterface import AbstractMachineInterface
from AbstractMachineInterface import DISCONNECTED, CONNECTED, BROKEN

#Statuses:
DOT_LENGTH = 20
BAD_BREAK_CHARS = ['[', '\033', 'm', ';', '0', '3', '1']
GOOD_BREAK_CHARS = ['.', '-', ' ', '\\', '/', '=', ')', ']', '_']
CONNECTION_TIMEOUT = 90 * 3600 #90 min
SSH_NEW_KEY = 'Are you sure you want to continue connecting'

class MachineInterface(AbstractMachineInterface):
    "Interface to a remote machine via pxssh"
    def __init__(self, machine_config, server_log):
        AbstractMachineInterface.__init__(self, machine_config, server_log)
        self.ssh_pass    = ''
        self.spkg_dir    = None # FIXME
        self.username    = None
        self.ip_address  = None
        self.platform    = None
        self.report_info = {}
        self.ssh_conn    = None
        self.set_data(machine_config)

    def set_data(self, machine_config):
        "Set up MachineInterface data"
        self.machine_name  = machine_config.machine_name
        self.server_home   = machine_config.server_home
        self.data          = machine_config.data
        self.username      = self.data.get("default_user", None)
        if self.username == None:
            if self.data.get("enc_username"):
                raise EnableRequiredException()
            msg = "'default_user' is not defined"
            raise IncompleteConfigurationException(self.machine_name, msg)
        self.ip_address    = self.data.get("ip_address", None)
        if self.ip_address == None:
            if self.data.get("enc_ip_address"):
                raise EnableRequiredException()
            msg = "'ip_address' is not defined"
            raise IncompleteConfigurationException(self.machine_name, msg)
        self.platform     = self.data.get("platform", None)
        if self.platform == None:
            if self.data.get("enc_platform"):
                raise EnableRequiredException()
            msg = "'platform' is not defined"
            raise IncompleteConfigurationException(self.machine_name, msg)

    def terminate(self):
        "Harshly disconnect from the remote system"
        status = OK
        if self.ssh_conn:
            result = self.ssh_conn.terminate(force=True)
            if result:
                self.status = DISCONNECTED
            else:
                status = FAIL
        self.unset_job()
        return status

    def connect(self):
        "Make a new connection to the remote system"
        self.ssh_conn = pxssh.pxssh()
        self.ssh_conn.timeout = 6000
        msg = "Connecting to %s..." % self.machine_name
        self.polling_log.debug(msg)
        #self.server_log.debug(msg, self.machine_name)
        try:
            self.server_log.debug(msg, self.machine_name)
            login_okay = self.ssh_conn.login(self.ip_address, self.username,
                                             self.ssh_pass, login_timeout=6000)
            if not login_okay:
                msg = "Could not connect."
                raise MachineUnavailableException(self.machine_name, msg)
            self.ssh_conn.sendline('stty -echo')
            self.ssh_conn.prompt()
        except (MachineUnavailableException, pexpect.TIMEOUT):
            msg = "SSH session failed on login."
            self.polling_log.error(msg)
            self.server_log.error(msg, self.machine_name)
            self.status = BROKEN
            return FAIL
        self.status = CONNECTED
        self.connect_time = time.time()
        return OK

    def control_c(self):
        #self.ssh_conn.sendcontrol('c')
        pass

    def freshen(self):
        '''Since we like to keep connections around for a long time,
        verify that our connection to the remote system is still working'''
        connection_age = time.time() - self.connect_time
        if self.status == DISCONNECTED or \
            connection_age > CONNECTION_TIMEOUT or \
            self.status == BROKEN:
            if self.status == CONNECTED:
                msg = "Assuming our connection to %s is stale after "\
                      "%4.2f minutes. Reconnecting..."
                msg = msg % (self.machine_name, connection_age / 60.0)
                self.polling_log.warning(msg)
                self.terminate()
            if self.connect() != OK:
                msg = "Unable to connect to %s." % self.machine_name
                raise MachineUnavailableException(self.machine_name, msg)

        dead = False
        try:
            self.ssh_conn.sendline('echo hello')
            if not self.ssh_conn.prompt(timeout = 5):
                dead = True
        except:
            dead = True

        if dead:
            msg = "Our connection handle is dead. Reconnecting..."
            self.polling_log.warning(msg)
            self.server_log.warning(msg, self.machine_name)
            try:
                self.terminate()
            except:
                pass
            if self.connect() != OK:
                msg = "Unable to connect to %s." % self.machine_name
                raise MachineUnavailableException(self.machine_name, msg)

    def _process_scp(self, scp_conn):
        "After we've started an scp, drive it until finished"
        expect_list = [pexpect.TIMEOUT, SSH_NEW_KEY,
                       '[pP]assword: ', 'Exit status']
        select_index = scp_conn.expect(expect_list, timeout=50)
        if select_index == 0:
            msg = scp_conn.before+'|'+scp_conn.after
            raise MachineUnavailableException(self.machine_name, msg)
        if select_index == 1:
            scp_conn.sendline('yes')
            scp_conn.expect('[pP]assword: ', timeout=30)
            select_index = scp_conn.expect([pexpect.TIMEOUT,
                                           '[pP]assword: '], timeout=50)
            if select_index == 0:
                msg = scp_conn.before+'|'+scp_conn.after
                raise MachineUnavailableException(self.machine_name, msg)
            scp_conn.sendline(self.ssh_pass)
        if select_index == 2:
            scp_conn.sendline(self.ssh_pass)
        if select_index == 3:
            pass
        scp_conn.expect(pexpect.EOF)
        scp_conn.close()
        return OK

    def get(self, dest_file):
        "secure copy from a remote host"
        self.polling_log.info( "Getting %s" % dest_file)
        cmd = 'scp -v %s@%s:%s .'
        cmd = cmd % (self.username, self.ip_address, dest_file)
        self.polling_log.debug("EXECUTING: %s" % cmd, cmd)
        scp_conn = pexpect.spawn(cmd, timeout=30)
        return self._process_scp(scp_conn)

    def scp(self, source, dest, verbose=True):
        "secure copy to a remote host"
        if not os.path.isfile(source):
            msg = "Attempting to send nonexistant file: %s" % source
            self.server_log.error(msg)
            self.polling_log.error(msg, self.machine_name)
            raise SecureCopyException(source, dest, msg)
        if verbose:
            msg = "Sending %s to %s:%s" % (source, self.machine_name, dest)
            self.polling_log.info(msg)
            self.server_log.info(msg, self.machine_name)
        else:
            msg = "Sending %s..." % (source)
            self.polling_log.debug(msg)
            self.server_log.debug(msg, self.machine_name)
        cmd = 'scp -v %s %s@%s:%s'
        cmd = cmd % (source, self.username, self.ip_address, dest)
        try:
            scp_conn = pexpect.spawn(cmd, timeout=600)
            select_index = scp_conn.expect([pexpect.TIMEOUT, SSH_NEW_KEY,
                                           '[pP]assword: ', 'Exit status'],
                                           timeout=600)
        except pexpect.EOF:
            msg = "Connection refused."
            raise MachineUnavailableException(dest, msg)
        if select_index == 0:
            msg = scp_conn.before+'|'+scp_conn.after
            raise MachineUnavailableException(dest, msg)
        if select_index == 1:
            scp_conn.sendline('yes')
            scp_conn.expect('[pP]assword: ', timeout=30)
            select_index = scp_conn.expect([pexpect.TIMEOUT,
                                           '[pP]assword: '], timeout=50)
            if select_index == 0:
                if type(scp_conn.before) == type("string") and \
                   type(scp_conn.after) == type("string"):
                    msg = scp_conn.before+'|'+scp_conn.after
                else:
                    msg = "before: (%s) after: (%s)"
                    msg = msg % (scp_conn.before, scp_conn.after)
                raise MachineUnavailableException(dest, msg)
            scp_conn.sendline(self.ssh_pass)
        if select_index == 2:
            self.polling_log.warning('Using password authentication')
            scp_conn.sendline(self.ssh_pass)
        if select_index == 3:
            pass
        scp_conn.expect(pexpect.EOF)
        scp_conn.close()
        return OK

    def scp_dict(self, copy_dict):
        "Use scp with a dictionary to copy files grouped by file type"
        self.connect()
        for file_desc in copy_dict:
            src, dst = copy_dict[file_desc]
            if not src.startswith('/'):
                src = os.path.join(self.server_home, src)
            msg = "SOURCE_PATH: %s // DEST_DIR: %s"
            msg =  msg % (src, dst)
            self.polling_log.debug(msg)
            status = self.scp(src, dst)

    def chdir(self, path):
        "Change the current directory on a remote session"
        if not path:
            path = self.spkg_dir
        self.polling_log.debug("Changing directory to %s" % path)
        self.ssh_conn.sendline ('cd %s' % path)
        self.ssh_conn.prompt()

    def gso(self, cmd, raise_on_error=True, cmd_debug=False):
        "Run a remote shell command"
        if cmd_debug:
            msg = "* RUNNING: %s" % cmd
            self.polling_log.debug(msg)
            self.server_log.debug(msg, self.machine_name)
        try:
            self.ssh_conn.sendline( cmd )
            self.ssh_conn.prompt()
        except pexpect.EOF:
            if raise_on_error:
                msg = "Error running %s" % (cmd)
                raise MachineUnavailableException(self.machine_name, msg)
            else:
                return ""
        output = self.ssh_conn.before.strip()
        if cmd_debug:
            self.polling_log.info("* OUTPUT: %s" % output)
        return output

    def run_cmd(self, command_string):
        "Run a remote shell command"
        self.report_info = ''
        #self.freshen(False) # Not sure why this was needed, but removing.
        return_code = OK
        self.ssh_conn.sendline(command_string)
        command_complete = False
        total_output = ''
        output_queue = ''
        output_checker = re.compile('(.*)'+self.ssh_conn.PROMPT)
        while True:
            output = self.ssh_conn.read_nonblocking()
            output_queue += output
            total_output += output
            output_queue = self.log_raw_data(output_queue)
            if output_checker.findall(total_output):
                #self.polling_log.info("EXITING")
                break
        #self.polling_log.info(output_queue)
        self.ssh_conn.setecho(False)
        return [OK, []]

    def check_result(self):
        "Examine the output of a command"
        self.ssh_conn.setecho(False)
        self.ssh_conn.sendline("echo $?")
        self.ssh_conn.prompt()
        return_code = 0
        try:
            return_code = int(str(self.ssh_conn.before.split()[0].strip()))
        except Exception:
            return FAIL
        return return_code

    def disconnect(self):
        self.connect_time = 0
        try:
            #self.ssh_conn.logout()
            self.ssh_conn.terminate(force=True)
        finally:
            self.status = DISCONNECTED


